//
//  ABCTKSmartCard.swift
//  ABC USB Demo
//
//  Copyright © 2025 AB Circle Limited. All rights reserved.
//

import CryptoTokenKit

/// (Rough) Helper class for exposing smart card reader functions from the Apple Crypto Token Kit
class ABCTKSmartCard {
    
    /// Return statuses
    enum Status : Error
    {
        /// Success
        case statusSuccess
        /// Smart Card Slot Manager is nil
        case statusSmartCardSlotManagerNil
        /// Smart Card Slot is nil
        case statusSmartCardSlotNil
        /// Smart Card Slot is missing
        case statusSmartCardSlotMissing
        /// Smart Card Slot is probing a card
        case statusSmartCardSlotProbingCard
        /// No card inserted in Smart Card Slot
        case statusNoCardInserted
        /// Card is not responding to commands
        case statusCardMute
        /// Card is not connected
        case statusCardNotConnected
        /// Card is already connected
        case statusCardAlreadyConnected
        /// Card connection failed
        case statusCardConnectionFailed
        /// Transmit failed
        case statusTransmitFailed
    }
    
    /// Represents a pool of SmartCard reader slots.
    private let slotManager = TKSmartCardSlotManager.default
    
    /// List of reader (slots)
    private var readerList = [String]()
    
    /// Current selected smart card slot
    private var smartCardSlot : TKSmartCardSlot? = nil
    
    /// Current connected smart card
    private var card : TKSmartCard? = nil
    
    /// Last status
    private var lastStatus: Status = Status.statusSuccess
    
    /// Flag to indicate if a card is connected or not
    private var connected: Bool = false
    
    /// Property returning list of reader slots
    var readers : [String] {
        get {
            return readerList
        }
    }
    
    /// Returns the card ATR if a card is connected
    var cardATR : Data? {
        get {
            if let card = card  {
                return card.slot.atr?.bytes
            }
            else {
                return nil
            }
        }
    }
    
    /// Returns the last status code
    var lastStatusCode: Status {
        get {
            return lastStatus
        }
    }
    
    /// Class initialiser
    init() {
        
    }
    
    /// Lists the AB Circle reader (slots) present on the system
    /// - Returns: Optional list of reader names (string)
    func listReaders() -> [String] {
        var slotNames: [String] = [""]
        
        readerList.removeAll()
        if let slotManager = slotManager {
            slotNames = slotManager.slotNames
            for slotName in slotNames {
                if slotName.contains("Circle CIR") {
                    readerList.append(slotName)
                }
            }
        }
        else {
            lastStatus = .statusSmartCardSlotManagerNil
        }
        
        return readerList
    }
    
    
    /// Connects to a card if present
    ///
    /// - Parameters:
    ///   - readerName: Name of current reader (slot)
    ///   - cardATR: Returned card ATR on success
    /// - Returns: statusSucces on success, else an error status
    func connect(readerName: String, cardATR: inout Data?) async -> Status {
        
        if (connected) {
            lastStatus = .statusCardAlreadyConnected
            return lastStatus
        }
        else if !getSmartCardSlot(readerName: readerName) {
            return lastStatus
        }
        else if !getSmartCard() {
            return lastStatus
        }
        else {
            do {
                if let card = card {
                    try connected = await card.beginSession()
                    if connected {
                        cardATR = card.slot.atr?.bytes
                    }
                }
            }
            catch {
                connected  = false
                lastStatus = .statusCardConnectionFailed
            }
        }
        
        return lastStatus
    }
    
    /// Disconnects a card if a card is present and connected
    /// - Returns: statusSuccess on success, else an error status
    func disconnect() -> Status {
        lastStatus = .statusSuccess
        if (connected) {
            if let card = card {
                card.endSession()
                connected = false
            }
            else
            {
                lastStatus = .statusNoCardInserted
            }
        }
        return lastStatus
    }
    
    
    /// Transmits an APDU to a connected card
    ///
    /// - Parameter apdu: APDU data to send
    /// - Returns: Data returned. Throws an error status on failure
    func transmit(apdu: Data) async throws -> Data {
        var response : Data = Data()
        
        lastStatus = .statusSuccess
        if connected {
            do {
                if let card = card {
                    try response = await card.transmit(apdu)
                }
            }
            catch {
                //todo use TKError code to determine the precice error
                lastStatus = .statusTransmitFailed
            }
        }
        else {
            lastStatus = .statusCardNotConnected
        }
        if lastStatus != .statusSuccess {
            throw lastStatus
        }
        return response
    }
    
    /// Function to retrieve the  smart slot associated by the entered readerName
    ///
    /// - Parameter readerName: Reader name
    /// - Returns: True on success, else false (also sets lastStatus).
    private func getSmartCardSlot(readerName: String) -> Bool {
        lastStatus = .statusSuccess
        
        guard let slotManager = slotManager else {
            lastStatus = .statusSmartCardSlotManagerNil
            return false
        }
        if smartCardSlot != nil {
            smartCardSlot = nil
        }
        smartCardSlot = slotManager.slotNamed(readerName)
        if (smartCardSlot == nil) {
            lastStatus = Status.statusSmartCardSlotNil
        }
        else if (smartCardSlot?.state == .missing) {
                lastStatus = .statusSmartCardSlotMissing
        }
        return (lastStatus == .statusSuccess)
    }
    
    /// Function to retrieve the card instance of a card in the current smart card reader slot
    ///
    /// - Returns: True on success, else false (also sets lastStatus).
    private func getSmartCard() -> Bool {
        lastStatus = .statusSuccess
        
        guard slotManager != nil else {
            lastStatus = .statusSmartCardSlotManagerNil
            return false
        }
        guard let smartCardSlot = smartCardSlot else {
            lastStatus = .statusSmartCardSlotManagerNil
            return false
        }
        if card != nil {
            card = nil
        }
        card = smartCardSlot.makeSmartCard()
        if card == nil {
            lastStatus = .statusNoCardInserted
        }
        return (lastStatus == .statusSuccess)
    }
}
