//
//  ViewController.swift
//  ABC USB Demo
//
//  Copyright © 2025 AB Circle Limited. All rights reserved.
//

import UIKit

/// Main view controller; controls the main screen
/// Also acts as delegate for the TextInputViewController and OptionsViewControllerProtocol
class ViewController: UIViewController, TextInputViewControllerProtocol {
    /// Enumerator holding the possible UI States
    private enum UIState {
        /// Initial (default) UI state on application startup
        case stateAppStart
        /// UI state when terminals are found
        case stateTerminalsFound
        /// UI state when connected to a terminal
        case stateConnected
    }
    
    /// Main instance of the AB Circle TKSmartCard class .
    private let abcTKSmartCard = ABCTKSmartCard()
    /// Global array of Card Terminals
    private var terminals  = [String]()
    /// Current selected Card Terminal
    private var currentTerminal: String = ""
    /// Flag indicating if connected to a Card or not
    private var connected: Bool = false
    /// Custom Card Terminal UIPickerView with Toolbar
    private let terminalPicker = ToolbarUIPickerView()
   
    /// Control outlet for custom key UIButton
    @IBOutlet weak var customKeyButton: UIButton!
    /// Control outlet for terminal UILabel
    @IBOutlet weak var labelTerminal: UILabel!
    /// Control outlet for terminal NoCaretUITextField
    @IBOutlet weak var selectedTerminal: NoCaretUITextField!
    /// Control outlet for Refresh UIButton
    @IBOutlet weak var buttonRefresh: UIButton!
    /// Control outlet for protocol UILabel
    @IBOutlet weak var labelProtocol: UILabel!
    /// Control outlet for protocol T=0 UISwitch
    @IBOutlet weak var protocolT0: UISwitch!
    /// Control outlet for protocol T=0 label
    @IBOutlet weak var labelT0: UILabel!
    /// Control outlet for protocol T=1 UISwitch
    @IBOutlet weak var protocolT1: UISwitch!
    /// Control outlet for protocol T=1 label
    @IBOutlet weak var labelT1: UILabel!
    /// Control outlet for Mode UILabel
    @IBOutlet weak var labelMode: UILabel!
    /// Control outlet for mode selection UISegmentedControl
    @IBOutlet weak var modeSelector: UISegmentedControl!
    /// Control outlet for connect UIButton
    @IBOutlet weak var buttonConnect: UIButton!
    /// Control outlet for log UIView
    @IBOutlet weak var viewLog: UIView!
    /// Control outlet for log UIScrollView
    @IBOutlet weak var scrollViewLog: UIScrollView!
    /// Control outlet for log NoCaretUITextView
    @IBOutlet weak var textViewLog: NoCaretUITextView!
    /// Control outlet for apdu label UILabel
    @IBOutlet weak var apduLabel: UILabel!
    /// Control outlet for apdu NoCaretUITextField
    @IBOutlet weak var apduTextField: NoCaretUITextField!
    /// Control outlet for audo send UIButton
    @IBOutlet weak var apduSendButton: UIButton!
    /// Control outlet for wait for change UIButton
    @IBOutlet weak var escapeCommandLabel: UILabel!
    /// Control outlet for escape command NoCaretUITextField
    @IBOutlet weak var escapeCommandTextField: NoCaretUITextField!
    /// Control outlet for escape command send UIButton
    @IBOutlet weak var escapeCommandSendButton: UIButton!
    
    /// Called after the controller's view is loaded into memory.
    /// Initialises the local variables, the controls and starts the terminal list updater
    override func viewDidLoad() {
        super.viewDidLoad()
        
        /// Initialise local variables
        currentTerminal = ""
        connected       = false

        /// Initialise terminal picker view
        selectedTerminal.inputView          = terminalPicker
        selectedTerminal.inputAccessoryView = terminalPicker.toolbar
        terminalPicker.delegate   = self
        terminalPicker.dataSource = self
        terminalPicker.toolbarDelegate = self
        terminalPicker.reloadAllComponents()
        /// Remove copy/paste buttons from toolbar on terminal picker
        let item : UITextInputAssistantItem = selectedTerminal.inputAssistantItem
        item.leadingBarButtonGroups = []
        item.trailingBarButtonGroups = []

        /// Initialise callbacks for showing APDU / Escape command input popup windows
        apduTextField.addTarget(self, action: #selector(showApduInputView), for: .touchDown)
        escapeCommandTextField.addTarget(self, action: #selector(showEscapeCommandInputView), for: .touchDown)
  
        /// Initialise log text view control
        initialiseTextView(textView: textViewLog)
       
        /// Initialise UI fields state
        enableUIFields(state: .stateAppStart)
    }
    
    /// Closes the APDU / Escape Command popup and retrieves its content
    ///
    /// - Parameters:
    ///   - textViewContent: Textview content from the text view on the popup
    ///   - isApdu: Flag indicating APDU popup or Escape Command popup
    func closePopup(textViewContent: String, isApdu: Bool) {
        if (isApdu) {
            apduTextField.text = textViewContent
        }
        else {
            escapeCommandTextField.text = textViewContent
        }
    }

    /// Callback handler for showing the APDU input view
    @objc private func showApduInputView() {
        TextInputViewController.showPopup(parentVC: self, textViewContent: apduTextField.text ?? "", isApdu: true)
    }

    /// Callback handler for showing the Escape Command input view
    @objc private func showEscapeCommandInputView() {
        TextInputViewController.showPopup(parentVC: self, textViewContent: escapeCommandTextField.text ?? "", isApdu: false)
    }
    
    /// Callback handler for handling connection mode change.
    /// Enables or disables the protocol fields.
    ///
    /// - Parameter sender: sender
    @IBAction func onModeChanged(_ sender: UISegmentedControl) {
        if selectedModeIsDirect() {
            enableProtocolFields(enabled: false)
        }
        else {
            enableProtocolFields(enabled: true)
        }
    }
    
    /// Callback handler for handling the Refresh button.
    /// Lists all  AB Circle USB readers (if any).
    ///
    /// - Parameter sender: sender
    @IBAction func onRefresh(_ sender: UIButton) {
        listReaders()
    }
    
    /// Updates the list of Card Terminals.
    /// Gets the latest readerlist  and checks if there are any changes.
    /// If there are changes, rebuilds the terminal picker and resets the terminal TextInput box if needed.
    private func listReaders() {
        DispatchQueue.global(qos: .userInteractive).async { [weak self] in
            guard let self = self else {
                return
            }
            let newTerminals: [String] = self.abcTKSmartCard.listReaders()
            if self.abcTKSmartCard.lastStatusCode != .statusSuccess {
                self.addLogLine(format: "Refresh failed (status: %@)", arguments: self.abcTKSmartCard.lastStatusCode as CVarArg)
            }
            if self.terminalsChanged(terminalList1: newTerminals, terminalList2: self.terminals) {
                self.terminals.removeAll()
                self.terminals = newTerminals
                DispatchQueue.main.async {
                    if self.terminals.count == 0 {
                        self.selectedTerminal.text = ""
                        if (self.connected) {
                            self.disconnect()
                        }
                        self.currentTerminal = ""
                        self.enableUIFields(state: .stateAppStart)
                    }
                    else {
                        self.enableUIFields(state: .stateTerminalsFound)
                    }
                    self.terminalPicker.selectRow(0, inComponent: 0, animated: false)
                    self.terminalPicker.reloadAllComponents()
                    if self.terminals.count != 0 {
                        self.selectCurrentTerminal(newTerminal: self.terminals[0])
                    }
                    else {
                        self.addLogLine(line: "No terminal selected")
                    }
                }
            }
            if self.terminals.count == 0 {
                self.addLogLine(line: "No terminals found")
            }
                
        }
    }
    
    /// Callback function handling the Protocol T=0 switch.
    /// Enables or disables the Connect button.
    ///
    /// - Parameter sender: sender
    @IBAction func onProtocolT0ValueChanged(_ sender: UISwitch) {
        if !protocolT0.isOn && !protocolT1.isOn {
            buttonConnect.isEnabled = false
        }
        else {
            buttonConnect.isEnabled = true
        }
    }
    
    /// Callback function handling the Protocol T=1 switch.
    /// Enables or disables the Connect button.
    ///
    /// - Parameter sender: sender
    @IBAction func onProtocolT1ValueChanged(_ sender: UISwitch) {
        if !protocolT0.isOn && !protocolT1.isOn {
            buttonConnect.isEnabled = false
        }
        else {
            buttonConnect.isEnabled = true
        }
    }
    
    /// Callback function handling the Connect button.
    /// If connected, tries to connect; else disconnects.
    ///
    /// - Parameter sender: sender
    @IBAction func onConnect(_ sender: Any) {
        Task { @MainActor in
            if connected {
                disconnect()
            }
            else {
                await connect()
            }
        }
    }
    
    /// Callback handler handling the "Send APDU" button.
    /// Sends an APDU command to a connected card (if any)
    ///
    /// - Parameter sender: sender
    @IBAction func onSendAPDU(_ sender: UIButton) {
        Task { @MainActor in
            if currentTerminal.isEmpty {
                addLogLine(line: "No terminal selected!")
                return
            }
            if !connected {
                addLogLine(line: "No card connected!")
                return
            }
            guard let apdu = apduTextField.text else {
                addLogLine(line: "Please enter an APDU first!")
                return
            }
            if apdu.isEmpty {
                addLogLine(line: "Please enter an APDU first!")
            }
            else {
                await sendAPDUThroughBasicChannel(apdu: apdu)
            }
        }
    }

    /// Callback handler handling the "Send Escape Command" button.
    /// Sends an Escape Command to a connected card (if any)
    ///  (Note: currently disabled)
    ///
    /// - Parameter sender: sender
    @IBAction func onSendEscapeCommand(_ sender: UIButton) {
        if currentTerminal.isEmpty {
            addLogLine(line: "No terminal selected!")
            return
        }
        if !connected {
            addLogLine(line: "No card connected!")
            return
        }
        guard let escapeCommand = escapeCommandTextField.text else {
            addLogLine(line: "Please enter a escape command first!")
            return
        }
        if escapeCommand.isEmpty {
            addLogLine(line: "Please enter a escape command first!")
        }
        else {
            sendEscapeCommand(escapeCommand: escapeCommand)
        }
    }
    
    /// Clears the log
    ///
    /// - Parameter sender: sender
    @IBAction func onClear(_ sender: UIButton) {
        textViewLog.text = ""
    }
    
    /// Initialises a text view control with empty text and a default border.
    ///
    /// - Parameter textView: Textview control to initialise
    private func initialiseTextView(textView: UITextView)
    {
        textView.text               = ""
        textView.layer.borderWidth  = 1.0
        textView.layer.cornerRadius = 8
        textView.layer.borderColor  =  UIColor.lightGray.cgColor
    }
    
    /// (Re-)Selects the current card terminal
    ///
    /// - Parameter newTerminal: New terminal selected on the UI
    private func selectCurrentTerminal(newTerminal: String)
    {
        if currentTerminal != newTerminal {
            currentTerminal = newTerminal
            selectedTerminal.text = currentTerminal
            addLogLine(format: "Terminal %@ selected", arguments: currentTerminal)
        }
    }
    
    /// Compares 2 Card Terminals array lists and determines if there are changes
    ///
    /// - Parameters:
    ///   - terminalList1: CardTerminal list 1
    ///   - terminalList2: CardTerminal list 2
    /// - Returns: True if changes were found, else false
    private func terminalsChanged(terminalList1: [String], terminalList2: [String]) -> Bool {
        return !terminalList1.elementsEqual(terminalList2, by: { $0 == $1 })
    }
    
    /// Sends an APDU through a card's basic channel and receives the response.
    ///
    /// - Parameter apdu: APDU in hexadecimal string form
    private func sendAPDUThroughBasicChannel(apdu: String) async {
        do {
            let apduData = Helper.hexStringToData(hexString: apdu)
            
            addLogLine(format: "APDU < %@", arguments: Helper.dataToHexString(data: apduData))
            let responseAPDU = try await abcTKSmartCard.transmit(apdu: apduData)
            addLogLine(format: "APDU > %@", arguments: Helper.dataToHexString(data: responseAPDU))
        }
        catch {
            
            addLogLine(format: "Send APDU failed with status %@", arguments: abcTKSmartCard.lastStatusCode as CVarArg)
        }
    }
    
  
    /// Sends an Escape command to a connected card
    /// (Note: currently disabled)
    ///
    /// - Parameter escapeCommand: Escape command in hexadecimal string form
    private func sendEscapeCommand(escapeCommand: String) {
        /*
        let command: Data = Helper.hexStringToData(hexString: escapeCommand)
        guard let card = card else {
            addLogLine(line: "Card not connected!")
            return
        }
        do {
            addLogLine(format: "ESC  < %@", arguments: escapeCommand)
            let response = try card.transmitControlCommand(controlCode: 3500, command: command)
            if let response = response {
                addLogLine(format: "ESC  > %@", arguments: Helper.dataToHexString(data: response))
            }
            else {
                addLogLine(line: "Send escape command success!")
            }
        }
        catch {
            addLogLine(format: "Send escape command failed with status %@, %@", arguments: card.lastStatusCode.name, card.lastStatusDescription)
        }
        */
    }
    
    /// Connects to a card in the selected terminal
    ///  (Note: ignores protocol at this moment)
    private func connect() async {
        //var displayProtocol: String
        //var cardProtocol: Card.ConnectionProtocol = .T0orT1
        var cardATR: Data? = nil
        
        if currentTerminal.isEmpty {
            addLogLine(line: "No terminal selected...")
            return
        }
        if (connected) {
            addLogLine(line: "Already connected...")
            return
        }
        /*
        if selectedModeIsDirect() {
            cardProtocol = .Direct
        }
        else {
            if protocolT0.isOn && !protocolT1.isOn {
                cardProtocol = .T0
            }
            else if !protocolT0.isOn && protocolT1.isOn {
                cardProtocol = .T1
            }
            else if protocolT0.isOn && protocolT1.isOn {
                cardProtocol = .T0orT1
            }
        }
        displayProtocol = cardProtocol.name
         
        addLogLine(format: "Connect using protocol %@", arguments: displayProtocol)
        */
        addLogLine(format: "Connecting card...")
        if await abcTKSmartCard.connect(readerName: currentTerminal, cardATR: &cardATR) == .statusSuccess {
            //addLogLine(format: "Success! Protocol: %@; ATR: %@", arguments: card.getProtocol(), card.getATR().toString())
            if let atr = cardATR {
                addLogLine(format: "Success! ATR: %@", arguments: Helper.dataToHexString(data: atr))
            }
            else {
                addLogLine(format: "Success!")
            }
            connected = true
            buttonConnect.setTitle("Disconnect", for: .normal)
            enableUIFields(state: .stateConnected)
        }
        else {
            addLogLine(format: "Connect failed!")
        }
    }
    
    /// Disconnects from a card (if connected)
    private func disconnect() {
        if (!connected) {
            addLogLine(line: "Not connected...")
            return
        }
        if abcTKSmartCard.disconnect() == .statusSuccess {
            connected = false
            buttonConnect.setTitle("Connect", for: .normal)
            enableUIFields(state: .stateTerminalsFound)
        }
        else {
            addLogLine(format: "Disconnect failed")
        }
    }
    
    /// Indicates if Direct Mode is selected
    ///
    /// - Returns: True if Direct Mode is selected, else false
    private func selectedModeIsDirect() -> Bool {
        return (modeSelector.selectedSegmentIndex == 1)
    }
    
    /// Enables groups of UI fields depending on the entered state
    ///
    /// - Parameter state: UI state (see enum UIState in this class for possible values)
    private func enableUIFields(state: UIState) {
        switch state {
            case .stateAppStart:
                enableMainFields(enabled: true)
                enableCommandFields(enabled: false)
                enableConnectionFields(enabled: false)
            case .stateTerminalsFound:
                enableMainFields(enabled: true)
                enableConnectionFields(enabled: true)
                enableCommandFields(enabled: false)
            case .stateConnected:
                enableMainFields(enabled: false)
                enableConnectionFields(enabled: false, excludeConnect: true)
                enableCommandFields(enabled: true)
        }
    }
    
    /// Enables or disables the main UI fields
    ///
    /// - Parameter enabled: True to set the fields as enabled, false to set the fields as disabled
    private func enableMainFields(enabled: Bool) {
        labelTerminal.isEnabled     = enabled
        selectedTerminal.isEnabled  = enabled
        buttonRefresh.isEnabled     = enabled
    }
    
    /// Enables or disables the command related UI fields
    ///
    /// - Parameter enabled: True to set the fields as enabled, false to set the fields as disabled
    private func enableCommandFields(enabled: Bool) {
        if enabled {
            if selectedModeIsDirect() {
                apduLabel.isEnabled             = false
                apduTextField.isEnabled         = false
                apduSendButton.isEnabled        = false
            }
            else {
                apduLabel.isEnabled             = true
                apduTextField.isEnabled         = true
                apduSendButton.isEnabled        = true
            }
            escapeCommandLabel.isEnabled        = true
            escapeCommandTextField.isEnabled    = true
            escapeCommandSendButton.isEnabled   = true
        }
        else {
            apduLabel.isEnabled                 = false
            apduTextField.isEnabled             = false
            apduSendButton.isEnabled            = false
            escapeCommandLabel.isEnabled        = false
            escapeCommandTextField.isEnabled    = false
            escapeCommandSendButton.isEnabled   = false
        }
    }
    
    /// Enables or disables the connection related UI fields
    ///
    /// - Parameter enabled: True to set the fields as enabled, false to set the fields as disabled
    private func enableConnectionFields(enabled: Bool, excludeConnect: Bool = false) {
        if !excludeConnect {
            buttonConnect.isEnabled = enabled
        }
        labelMode.isEnabled             = enabled
        modeSelector.isEnabled          = enabled
        enableProtocolFields(enabled: enabled)
    }
    
    /// Enables or disables the protocol selection related UI fields
    ///
    /// - Parameter enabled: True to set the fields as enabled, false to set the fields as disabled
    private func enableProtocolFields(enabled: Bool) {
        if enabled {
            if selectedModeIsDirect() {
                labelProtocol.isEnabled = false
                labelT0.isEnabled       = false
                labelT1.isEnabled       = false
                protocolT0.isEnabled    = false
                protocolT1.isEnabled    = false
            }
            else {
                labelProtocol.isEnabled = true
                labelT0.isEnabled       = true
                labelT1.isEnabled       = true
                protocolT0.isEnabled    = true
                protocolT1.isEnabled    = true
            }
        }
        else {
            labelProtocol.isEnabled = false
            labelT0.isEnabled       = false
            labelT1.isEnabled       = false
            protocolT0.isEnabled    = false
            protocolT1.isEnabled    = false
        }
    }
    
    /// Output the terminal list to the log ui field
    private func outputTerminalsToLog() {
        for terminal in terminals {
            addLogLine(format: "Terminal %@ found", arguments: terminal)
        }
    }
    
    /// Add a formatted line to the log ui field
    ///
    /// - Parameters:
    ///   - format: Line format to use
    ///   - arguments: Arguments for the format
    private func addLogLine(format : String, arguments: CVarArg...) {
        let message = String(format: format.replacingOccurrences(of: "%s", with: "%@"), arguments: arguments)

        self.addLogLine(line: message)
        self.scrollLogToEnd()
    }
    
    /// Adds a line to the log ui filed
    ///
    /// - Parameter line: Line to add
    private func addLogLine(line: String) {
        DispatchQueue.main.async {
            let formatter = DateFormatter()
            
            formatter.dateFormat = "HH:mm:ss"
            let message = "[" + formatter.string(from: Date()) + "] " + line + "\n"

            self.textViewLog.text = self.textViewLog.text + message
        }
    }
    
    /// Scrolls the log ui field contents to the end
    private func scrollLogToEnd() {
        DispatchQueue.main.async {
            let range = NSRange(location: self.textViewLog.text.count, length: 0)
            self.textViewLog.scrollRangeToVisible(range)
        }
    }
}

/// ViewController extension acting as  UIPickerViewDelegate and UIPickerViewDataSource
extension ViewController : UIPickerViewDelegate, UIPickerViewDataSource {
    
    /// Gets the number of components for the picker view. Here always 1, since we only use 1 component.
    ///
    /// - Parameter pickerView: UIPickerView to handle
    /// - Returns: Int holding the number of components
    func numberOfComponents(in pickerView: UIPickerView) -> Int {
        return 1
    }
    
    /// Returns the number of rows for a component.
    /// Here returns the number of rows in the main CardTerminal list (as that is the data source for the UIPickerView)
    ///
    /// - Parameters:
    ///   - pickerView: An object representing the picker view requesting the data.
    ///   - component: A zero-indexed number identifying a component of pickerView. Components are numbered left-to-right.
    /// - Returns: Int holding the number of rows for the specified component
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int {
        return terminals.count
    }
    
    /// Called by the picker view when it needs the title to use for a given row in a given component.
    /// Here returns the name of the terminal from the terminal list associated with the row as datasource
    ///
    /// - Parameters:
    ///   - pickerView: An object representing the picker view requesting the data.
    ///   - row: A zero-indexed number identifying a row of component. Rows are numbered top-to-bottom.
    ///   - component: A zero-indexed number identifying a component of pickerView. Components are numbered left-to-right.
    /// - Returns: The string to use as the title of the indicated component row.
    func pickerView(_ pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {
        if row < terminals.count {
            return terminals[row]
        }
        else {
            return nil
        }
    }
}

/// ViewController extension acting as ToolbarUIPickerViewDelegate
extension ViewController: ToolbarUIPickerViewDelegateProtocol {
    
    /// User tapped 'Done' button. Here selects the current terminal and closes the UIPickerView.
    func didTapDone() {
        let row = terminalPicker.selectedRow(inComponent: 0)
        
        if row < terminals.count {
            terminalPicker.selectRow(row, inComponent: 0, animated: false)
      
            selectCurrentTerminal(newTerminal: terminals[row])
        }
        selectedTerminal.resignFirstResponder()
    }
    
    /// User tapped 'Cancel' button. Closes the UIPickerView.
    func didTapCancel() {
        selectedTerminal.resignFirstResponder()
    }
}

